package org.steveshipway.blueprints;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.Logger;
import org.steveshipway.blueprints.directive.*;
import org.steveshipway.blueprints.lib.Constants;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.world.World;
import net.minecraft.world.chunk.Chunk;


// Minecraft 1.7
//import org.steveshipway.blueprints.util17.*;
// Minecraft 1.8
import org.steveshipway.blueprints.util18.*;

import net.minecraft.util.BlockPos;

public class Blueprint {
	private boolean floaty = false;  // float y coordinate before build
	private boolean dry = false;     // float on water
	private boolean recurse = false; // allow recursion
	public boolean visible = true;   // hidden or visible blueprint
	private LinkedList steps = null; // build steps
	public String name;              // my name
	public String helptext = "No description available";
	public float auto_chance = 0;
	public Set biomes = new HashSet(); 
	private Logger logger = null;
	private HashSet<Block> softlist = new HashSet<Block>();
	
	public Blueprint(String name, String description, Boolean v, LinkedList steps, Logger logger) {
		// create this blueprint
		this.name = name;
		this.steps = steps;
		this.visible = v;
		this.logger = logger;
		if( description != null && description != "" ) {
			this.helptext = description;
		}
		// set up the default softblock list 
		this.softlist.add(Blocks.air);
		this.softlist.add(Blocks.water);
		this.softlist.add(Blocks.flowing_water);
		
		if( steps.isEmpty() ) {
        	logger.warn("["+name+"] Created a blueprint with no steps?");
        	return;
		}
		// iterate over the commands, removing any options and setting the flags
		Iterator i = this.steps.iterator();
	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	        String s = d.description();
	        if( s != null ) {
	        	this.helptext = s;
	        	// this.steps.remove(d);
	        	continue;
	        }
	        s = d.option();
	        if( s != null ) {
	        	if( s.contains("float")) { this.floaty = true; }
	        	if( s.contains("sink")) { this.floaty = false; }
	        	if( s.contains("dry")) {
	        		this.dry = true;
	        		this.floaty = true;
	        	}
	        	if( s.contains("recurs")) { // catch recurse and recursive
	        		this.recurse = true;
	        	}
	        	if( s.startsWith("soft") ) {
	        		this.softlist.clear();
	        		for(String blockname : s.split("[\\s,]+") ) {
	        			if( ! blockname.contains(":") ) { blockname = "minecraft:"+blockname; }
	        			Block b = Block.getBlockFromName(blockname);
	        			if( b == null ) {
	        				logger.warn("Invalid block name in OPTION SOFT directive: "+blockname);
	        			} else {
	        				this.softlist.add(b);
	        			}
	        		}
	        	}

	        	// this.steps.remove(d);
	        	continue;
	        }
	        float f = d.auto_chance();
	        if( f > 0 && f <= 1 ) {
	        	this.auto_chance = f;
	        	this.biomes = ((AutoDirective)d).biomes;
	        	logger.info("Added an auto-generated plan '"+this.name+"', chance "+Math.floor(this.auto_chance*Constants.CHANCE_BASE)+"/"+Constants.CHANCE_BASE);
	        	continue;
	        } else if( f > 1 ) {
	        	logger.info("Invalid auto chance "+this.auto_chance);
	        }
	    }
    	//logger.info("Created a blueprint with "+this.steps.size()+" steps");
	}
	
	public boolean isVisible() {
		return visible;
	}
	public boolean setVisible(boolean v) {
		visible = v;
		return visible;
	}

	public String helpText() {
		if( this.helptext != null ) {
			return this.helptext;
		}
		return "No description available";
	}
	public String name() { return this.name; }
	
	// build this blueprint, at the specified coordinates.
	public boolean build(World world, BlockPos location, int r ) {
		// iterate over steps, actioning each in turn
		// may require (gulp) recursion...
		// return true if it was successfully built
		boolean rv = true;
		if( world.isRemote ) {
        	logger.warn("Blueprint build called with remote world!  Should never happen.");
			return false;			
		}
		if( this.steps == null || this.steps.size() == 0 ) {
        	logger.warn("Build of '"+this.name+"' had no steps!  This seems wrong...");
			return false;
		}
    	logger.info("Build of blueprint '"+this.name+"' starting: "+this.steps.size()+" steps");
		
    	// floating option -- go up.down to find surface of world
    	if( this.floaty ) {
    		BlockPos blk = new BlockPos(location);
    		Set<Block> floatthrough = new HashSet<Block>();
    		
    		floatthrough.add(Blocks.air); // treat air as air
    		floatthrough.add(Blocks.log); // treat trees as air
    		floatthrough.add(Blocks.log2);
    		floatthrough.add(Blocks.leaves);
    		floatthrough.add(Blocks.leaves2);
    		floatthrough.add(Blocks.yellow_flower); // treat flowers and plants as air
    		floatthrough.add(Blocks.red_flower);
    		floatthrough.add(Blocks.double_plant);
    		floatthrough.add(Blocks.snow);
    		floatthrough.add(Blocks.wheat);
    		floatthrough.add(Blocks.pumpkin);
    		floatthrough.add(Blocks.melon_block);
    		floatthrough.add(Blocks.pumpkin_stem);
    		floatthrough.add(Blocks.melon_stem);
    		floatthrough.add(Blocks.waterlily);
    		floatthrough.add(Blocks.reeds);
    		floatthrough.add(Blocks.cactus);
    		floatthrough.add(Blocks.potatoes);
    		floatthrough.add(Blocks.rail); // treat decorations as air
    		floatthrough.add(Blocks.torch);
    		floatthrough.add(Blocks.redstone_wire);
    		floatthrough.add(Blocks.wall_banner);
    		floatthrough.add(Blocks.standing_banner);
    		floatthrough.add(Blocks.carpet);
    		floatthrough.add(Blocks.skull);
    		floatthrough.add(Blocks.tripwire);
    		floatthrough.add(Blocks.fire);
    		if( ! this.dry ) {  // treat liquids as air
        		floatthrough.add(Blocks.lava);
        		floatthrough.add(Blocks.flowing_lava);
        		floatthrough.add(Blocks.water);
        		floatthrough.add(Blocks.flowing_water);    			
    		}
    		// float down to ground
    		blk = location.down();
    		Block b = Util.getBlock(world,blk);
    		while( blk.getY() > 0 && floatthrough.contains(b) ) {
    			location = location.down();
    			blk = location.down();
    			b = Util.getBlock(world,blk);
    		}
    		// float up to air
       		while( blk.getY() < 128 && ! floatthrough.contains(b) ) {
    			location = location.up(); 
       			blk = location.up();
       			b = Util.getBlock(world,blk);
    		}
        	logger.info("Floating Y coordinate to surface level: "+location.getY());
    	}
    	
		Iterator i = this.steps.iterator();
	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	    	//CheshireCat.logger.debug("Calling step: "+d.asText());
	        rv = d.build(world, location, r, this.logger,this.softlist);
	        if(!rv) {
	        	logger.error("Build of '"+this.name+"' failed with step: ["+d.asText()+"]");
	        	break; 
	        }
	    }
    	//CheshireCat.logger.info("Build of blueprint '"+this.name+"' ended");
		return rv;
	}
	public String asText() {
		String s = "";
		Iterator i = this.steps.iterator();
		if( visible ) {
			s = "PUBLIC";
		} else {
			s = "PRIVATE";
		}
		s = s + " PLAN " + name + " " + helptext + "\n";

	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	        s = s + "  " + d.asText() + "\n";
	    }
	    s = s + "END\n";
		return s;
	}
	public boolean save(String f) {
		boolean rv = true;
		if( f == null ) { return false; }
		Writer writer = null;
		try {
		    writer = new BufferedWriter(new OutputStreamWriter(
		          new FileOutputStream(f), "utf-8")
		    );
		    writer.write("# Blueprint for: "+this.name+"\n");
		    writer.write("# "+this.helptext+"\n");
		    writer.write("# "+this.steps.size()+" steps.\n");
			writer.write(this.asText()+"\n");
		    writer.write("# End of blueprint\n");
		} catch (IOException ex) {
			logger.error("Unable to write file "+f+": "+ex.getMessage());
			rv = false;
		} finally {
		    try {writer.close();} catch (Exception ex) {}
		}
		return rv;
	}
}
